#include "LightweightDatabaseServer.h"
#include "LightweightDatabaseClient.h"
#include "TableSerializer.h"
#include "BitStream.h"
#include "StringCompressor.h"
#include "DS_Table.h"
#include "LightweightDatabaseClient.h"
#include "LightweightDatabaseServer.h"
#include "RakPeerInterface.h"
#include "RakNetworkFactory.h"
#include "RakSleep.h"
#include "MessageIdentifiers.h"
#include "NatPunchthrough.h"
#include "MasterServer.h"

#ifdef WIN32
#include <stdio.h>
#include <time.h>
#include <stdarg.h>
#else
#include <curses.h>
#include <stdlib.h>
#include <string>
#include <map>
#endif

void print_log(const char* format, ...)
{
	time_t rawtime;
	struct tm * timeinfo;
	time ( &rawtime );
	timeinfo = localtime ( &rawtime );
	printf("%02d-%02d-%d %02d:%02d:%02d\t",timeinfo->tm_mday, 1+timeinfo->tm_mon, 1900+timeinfo->tm_year, timeinfo->tm_hour, timeinfo->tm_min, timeinfo->tm_sec);

	va_list va;
	va_start( va, format );
	vprintf(format, va);
}

LightweightDatabaseServer databaseServer;
NatPunchthrough natPunchthrough;

void RemoveTable(std::string table)
{
	if (table == "") return;	
	if (databaseServer.RemoveTable(const_cast<char*>(table.c_str())))
		print_log("Success\n");
	else
		print_log("Table %s not found\n", table.c_str());
}

void ProcessPacket(Packet *packet, char* caller)
{
	switch (packet->data[0])
	{
		case ID_DISCONNECTION_NOTIFICATION:
			print_log("%s: %s has diconnected\n", caller, packet->systemAddress.ToString());
			break;
		case ID_CONNECTION_LOST:
			print_log("%s: Connection to %s lost\n", caller, packet->systemAddress.ToString());
			break;
		case ID_NO_FREE_INCOMING_CONNECTIONS:
			print_log("%s: No free incoming connection for %s\n", caller, packet->systemAddress.ToString());
			break;
		case ID_NEW_INCOMING_CONNECTION:
			print_log("%s: New connection established to %s\n", caller, packet->systemAddress.ToString());
			break;
		case ID_CONNECTION_REQUEST_ACCEPTED:
			print_log("%s: Connection to %s accepted\n", caller, packet->systemAddress.ToString());
			break;
		case ID_CONNECTION_ATTEMPT_FAILED:
			print_log("%s: Connection attempt to %s failed\n", caller, packet->systemAddress.ToString());
			break;
		case ID_DATABASE_UNKNOWN_TABLE:
			print_log("%s: ID_DATABASE_UNKNOWN_TABLE to %s\n", caller, packet->systemAddress.ToString());
			break;
		case ID_DATABASE_INCORRECT_PASSWORD:
			print_log("%s: ID_DATABASE_INCORRECT_PASSWORD to %s\n", caller, packet->systemAddress.ToString());
			break;
		case ID_DATABASE_QUERY_REPLY:
			print_log("%s: ID_DATABASE_QUERY_REPLY to %s\n", caller, packet->systemAddress.ToString());
			break;
		case ID_NAT_TARGET_NOT_CONNECTED:
		{
			SystemAddress systemAddress;
			RakNet::BitStream b(packet->data, packet->length, false);
			b.IgnoreBits(8); // Ignore the ID_...
			b.Read(systemAddress);
			print_log("%s: ID_NAT_TARGET_NOT_CONNECTED to %s\n", caller, systemAddress.ToString());
			break;
		}
		case ID_NAT_TARGET_CONNECTION_LOST:
		{
			SystemAddress systemAddress;
			RakNet::BitStream b(packet->data, packet->length, false);
			b.IgnoreBits(8); // Ignore the ID_...
			b.Read(systemAddress);
			print_log("%s: ID_NAT_TARGET_CONNECTION_LOST to %s\n", caller, systemAddress.ToString());
			break;
		}
		default:
			print_log("%s: Unknown ID %d from %s\n", caller, packet->data[0], packet->systemAddress.ToString());
	}
}

int main(int argc, char *argv[])
{  
#ifndef WIN32
	setlinebuf(stdout);
#endif

	RakPeerInterface *masterPeer = RakNetworkFactory::GetRakPeerInterface();
	RakPeerInterface *natPeer = RakNetworkFactory::GetRakPeerInterface();	// The facilitator
		
	int port;
	if (argc > 1)
		port = atoi(argv[1]);
	else
		port = 12345;
	print_log("Master server port set to %d\n",port); 
	int facilitatorPort;
	if (argc > 2)
		facilitatorPort = atoi(argv[2]);
	else
		facilitatorPort = 50000;
	print_log("Facilitator port set to %d\n",facilitatorPort); 
		
	natPunchthrough.FacilitateConnections(true);

	// Start the peers
	masterPeer->SetMaximumIncomingConnections(1000);
	SocketDescriptor sd(port,0);
	masterPeer->Startup(1000, 30, &sd, 1);
	masterPeer->AttachPlugin(&databaseServer);
	
	natPeer->SetMaximumIncomingConnections(1000);
	sd.port = facilitatorPort;
	natPeer->Startup(1000, 30, &sd, 1);
	natPeer->AttachPlugin(&natPunchthrough);

	Packet *p;
	while (1)
	{
		p=masterPeer->Receive();
		while (p)
		{
			ProcessPacket(p, "MasterServer");
			masterPeer->DeallocatePacket(p);
			p=masterPeer->Receive();
		}
		p=natPeer->Receive();
		while (p)
		{
			ProcessPacket(p, "Facilitator");
			natPeer->DeallocatePacket(p);
			p=natPeer->Receive();
		}
		RakSleep(30);
	}

	masterPeer->Shutdown(100,0);
	natPeer->Shutdown(100,0);
	RakNetworkFactory::DestroyRakPeerInterface(masterPeer);
	RakNetworkFactory::DestroyRakPeerInterface(natPeer);
			
	return 0;
}
